@@ -1,4 +1,7 @@
+from copy import copy
+from dataclasses import dataclass
 from functools import wraps
+from time import time
 from typing import Optional, List, Any, Iterable, Dict, Sequence
 import logging
 
@@ -6,7 +9,7 @@ from .core.values.value_mapping import get_map_helper
 from .tdm_builder.tdm_builder import AbstractTdmBuilder
 from .providers.gql_providers import AbstractGQLClient
 from .schema.api_schema import Query, Mutation, StoryPagination, ConceptLinkPagination, ConceptFactPagination, \
-    ConceptPropertyPagination, SortDirection, ConceptSorting, DocumentSorting, DocumentGrouping
+    ConceptPropertyPagination, SortDirection, ConceptSorting, DocumentSorting, DocumentGrouping, TimestampInterval
 from .schema.api_schema import \
     ValueInput, IntValueInput, StringValueInput, DateTimeValue, StringLocaleValue, StringValue, LinkValue, \
     DoubleValue, IntValue, ConceptPropertyFilterSettings, ConceptLinkFilterSettings, ExtraSettings, Story, Document,\
@@ -23,6 +26,13 @@ from .schema.crawlers_api_schema import Crawler, CrawlerPagination
 from .schema import utils_api_schema as uas
 from sgqlc.operation import Operation, Fragment
 
+
+@dataclass
+class KBIteratorConfig:
+    max_total_count: int
+    earliest_created_time: int
+
+
 logger = logging.getLogger(__name__)
 
 
@@ -39,22 +49,31 @@ def check_utils_gql_client(f):
 class TalismanAPIAdapter:
     def __init__(
         self, gql_client: Optional[AbstractGQLClient], types: Dict, tdm_builder: AbstractTdmBuilder = None,
-        utils_gql_client: Optional[AbstractGQLClient] = None
+        utils_gql_client: Optional[AbstractGQLClient] = None, kb_iterator_config: KBIteratorConfig = None,
+        limit: int = 100
     ) -> None:
         self._gql_client = gql_client
         self._utils_gql_client = utils_gql_client
         self._types = types
         self._cache = {}
+        self._limit = limit
 
         self.document_fields_truncated = ('id', 'external_url')
         self.document_fields = (
             'id', 'title', 'external_url', 'publication_author', 'publication_date', 'internal_url', 'markers'
         )
+        self.document_fields_extended = (
+            'id', 'title', 'publication_author', 'publication_date', 'external_url',
+            'markers', 'notes', 'access_level', 'trust_level'
+        )
 
         self.document_text_fields_truncated = ('text',)
         self.document_text_fields = ('node_id', 'text')
         self.document_text_metadata_fields = ('paragraph_type',)
 
+        self.document_platform_fields = ('id',)
+        self.document_account_fields = ('id',)
+
         self.concept_fields = (
             'id', 'name', 'notes', 'metric', 'markers', 'system_registration_date', 'system_update_date'
         )
@@ -73,12 +92,6 @@ class TalismanAPIAdapter:
         self.concept_link_type_fields_truncated = ('id', 'name', 'is_directed')
 
         self.concept_fact_fields = ('id',)
-        self.concept_fact_document_fields = (
-            'id', 'title', 'publication_author', 'publication_date', 'external_url',
-            'markers', 'notes', 'access_level', 'trust_level'
-        )
-        self.concept_fact_document_platform_fields = ('id',)
-        self.concept_fact_document_account_fields = ('id',)
 
         self.composite_concept_widget_type = ('id', 'name')
         self.composite_concept_widget_type_columns_info = ('name',)
@@ -88,9 +101,25 @@ class TalismanAPIAdapter:
 
         self.tdm_builder = tdm_builder
 
+        if kb_iterator_config:
+            self.kb_iterator_config = kb_iterator_config
+        else:
+            self.kb_iterator_config = KBIteratorConfig(1000, 1609448400)  # Fri Jan 01 2021 00:00:00 GMT+0300
+
     def get_tdm_builder(self) -> AbstractTdmBuilder:
         return self.tdm_builder
 
+    @property
+    def limit(self):
+        return self._limit
+
+    @limit.setter
+    def limit(self, new_limit: int):
+        self._limit = new_limit
+
+    def get_take_value(self, take: Optional[int]) -> int:
+        return self.limit if take is None else take
+
     def _configure_property_value_type_fields(self, graphql_value, truncated: bool = True):
         conpvt_frag: Fragment = Fragment(ConceptPropertyValueType, 'ConceptPropertyValueType')
         for f in self.cpvt_fields_truncated if truncated else self.cpvt_fields:
@@ -138,9 +167,8 @@ class TalismanAPIAdapter:
 
     def _configure_output_concept_fields(
         self, concept_object, with_aliases=False, with_properties=False, with_links=False,
-        with_link_properties=False
+        with_link_properties=False, with_facts=False
     ):
-
         concept_object.__fields__(*self.concept_fields)
         concept_object.concept_type.__fields__(*self.concept_type_fields)
         if with_aliases:
@@ -163,22 +191,38 @@ class TalismanAPIAdapter:
                 limit=10000,
                 filter_settings=ConceptLinkFilterSettings()
             )
-            lcl = pcl.list_concept_link()
-            lcl.__fields__(*self.concept_link_fields)
-            lcl.concept_from().__fields__(*self.concept_link_concept_from_fields)
-            lcl.concept_to().__fields__(*self.concept_link_concept_to_fields)
-            lcl.concept_link_type().__fields__(*self.concept_link_type_fields_truncated)
+            self._configure_output_link_fields(pcl.list_concept_link(), with_link_properties=with_link_properties)
+        if with_facts:
+            pcf: ConceptFactPagination = concept_object.pagination_concept_fact(
+                offset=0,
+                limit=10000,
+                filter_settings=DocumentLinkFilterSetting()
+            )
+            lcf = pcf.list_concept_fact()
+            lcf.__fields__(*self.concept_fact_fields)
+            d = lcf.document()
+            d.__fields__(*self.document_fields_extended)
+            dm = d.metadata()
+            dm.platform().__fields__(*self.document_platform_fields)
+            dm.account().__fields__(*self.document_account_fields)
 
-            if with_link_properties:
-                pcp: ConceptPropertyPagination = lcl.pagination_concept_link_property(
-                    offset=0,
-                    limit=10000,
-                    filter_settings=ConceptPropertyFilterSettings()
-                )
-                lcp = pcp.list_concept_property()
-                lcp.__fields__(*self.concept_property_fields)
-                lcp.property_type().__fields__(*self.concept_property_type_fields)
-                self._configure_output_value_fields(lcp.value)
+    def _configure_output_link_fields(
+        self, link_object, with_link_properties=False
+    ):
+        link_object.__fields__(*self.concept_link_fields)
+        link_object.concept_from().__fields__(*self.concept_link_concept_from_fields)
+        link_object.concept_to().__fields__(*self.concept_link_concept_to_fields)
+        link_object.concept_link_type().__fields__(*self.concept_link_type_fields_truncated)
+        if with_link_properties:
+            pcp: ConceptPropertyPagination = link_object.pagination_concept_link_property(
+                offset=0,
+                limit=10000,
+                filter_settings=ConceptPropertyFilterSettings()
+            )
+            lcp = pcp.list_concept_property()
+            lcp.__fields__(*self.concept_property_fields)
+            lcp.property_type().__fields__(*self.concept_property_type_fields)
+            self._configure_output_value_fields(lcp.value)
 
     def _create_concept_with_input(
         self, input: ConceptMutationInput, with_properties=False, with_links=False, with_link_properties=False
@@ -204,41 +248,67 @@ class TalismanAPIAdapter:
         return res.add_concept
 
     def get_all_documents(
-        self, grouping: DocumentGrouping = 'none', query: str = None,
-        filter_settings: DocumentFilterSettings = None, direction: SortDirection = 'descending',
-        sort_field: DocumentSorting = 'score', extra_settings: ExtraSettings = None
+        self, grouping: DocumentGrouping = 'none', filter_settings: DocumentFilterSettings = None,
+        direction: SortDirection = 'descending', sort_field: DocumentSorting = 'score',
+        extra_settings: ExtraSettings = None, with_extended_information: bool = False
     ) -> Iterable[Story]:
         if filter_settings is None:
             filter_settings = DocumentFilterSettings()
         if extra_settings is None:
             extra_settings = ExtraSettings()
+
+        total = self.get_documents_count(filter_settings=filter_settings)
+
+        if total > self.kb_iterator_config.max_total_count:
+            had_creation_date = hasattr(filter_settings, 'registration_date')
+            old_timestamp_interval = None
+            if had_creation_date:
+                old_timestamp_interval = copy(filter_settings.registration_date)
+            start: int = getattr(old_timestamp_interval, 'start', self.kb_iterator_config.earliest_created_time)
+            end: int = getattr(old_timestamp_interval, 'end', int(time()))
+            middle: int = (end + start) // 2
+
+            for start, end in (start, middle), (middle, end):
+                filter_settings.registration_date = TimestampInterval(start=start, end=end)
+                for c in self.get_all_documents(
+                    grouping=grouping, filter_settings=filter_settings, direction=direction, sort_field=sort_field,
+                    extra_settings=extra_settings, with_extended_information=with_extended_information
+                ):
+                    yield c
+
+            if had_creation_date:
+                filter_settings.registration_date = old_timestamp_interval
+            else:
+                delattr(filter_settings, 'registration_date')
+            return
+        elif not total:
+            return
+
         documents: Iterable = [None]
         i: int = 0
         while documents:
             documents = self.get_documents(
-                skip=i * 20, take=20, grouping=grouping, query=query,
-                filter_settings=filter_settings, direction=direction,
-                sort_field=sort_field, extra_settings=extra_settings
+                skip=i * self._limit, take=self._limit, grouping=grouping, filter_settings=filter_settings,
+                direction=direction, sort_field=sort_field, extra_settings=extra_settings,
+                with_extended_information=with_extended_information
             )
             for d in documents:
                 yield d
             i += 1
 
     def get_documents(
-        self, skip: int = 0, take: int = 20, grouping: DocumentGrouping = 'none', query: str = None,
-        filter_settings: DocumentFilterSettings = None,
-        direction: SortDirection = 'descending', sort_field: DocumentSorting = 'score',
-        extra_settings: ExtraSettings = None
+        self, skip: int = 0, take: Optional[int] = None, grouping: DocumentGrouping = 'none',
+        filter_settings: DocumentFilterSettings = None, direction: SortDirection = 'descending',
+        sort_field: DocumentSorting = 'score', extra_settings: ExtraSettings = None,
+        with_extended_information: bool = False
     ) -> Sequence[Story]:
+        take = self.get_take_value(take)
         pagination_story_kwargs = dict()
         if filter_settings is None:
             filter_settings = DocumentFilterSettings()
         if extra_settings is None:
             extra_settings = ExtraSettings()
 
-        if query:
-            pagination_story_kwargs['query'] = query
-
         op = Operation(Query)
         ps: StoryPagination = op.pagination_story(
             offset=skip,
@@ -252,54 +322,45 @@ class TalismanAPIAdapter:
         )
         ps.list_story().list_document().__fields__(*self.document_fields_truncated)
         m = ps.list_story().main()
-        m.__fields__(*self.document_fields_truncated)
+        if with_extended_information:
+            m.__fields__(*self.document_fields_extended)
+            mdm = m.metadata()
+            mdm.platform().__fields__(*self.document_platform_fields)
+            mdm.account().__fields__(*self.document_account_fields)
+        else:
+            m.__fields__(*self.document_fields_truncated)
         m.text().__fields__(*self.document_text_fields_truncated)
 
         res = self._gql_client.execute(op)
         res = op + res
         return res.pagination_story.list_story
 
-    def get_documents_count(self) -> int:
+    def get_documents_count(self, filter_settings: DocumentFilterSettings = None) -> int:
+        if filter_settings is None:
+            filter_settings = DocumentFilterSettings()
+
         op = Operation(Query)
         ps: StoryPagination = op.pagination_story(
             limit=1,
+            filter_settings=filter_settings,
             extra_settings=ExtraSettings()
         )
-        ps.total()
+        ps.show_total()
         res = self._gql_client.execute(op)
         res = op + res
-        return res.pagination_story.total
+        return res.pagination_story.show_total
 
     def get_documents_by_limit_offset_filter_extra_settings(
-        self, skip: int, take: int, filter_settings: DocumentFilterSettings, extra_settings: ExtraSettings
+        self, skip: int = 0, take: Optional[int] = None, filter_settings: Optional[DocumentFilterSettings] = None,
+        extra_settings: Optional[ExtraSettings] = None
     ) -> Sequence[Story]:
-
+        take = self.get_take_value(take)
         op = Operation(Query)
         ps: StoryPagination = op.pagination_story(
             offset=skip,
             limit=take,
-            extra_settings=extra_settings,
-            filter_settings=filter_settings
-        )
-        ps.list_story().list_document().__fields__(*self.document_fields_truncated)
-        m = ps.list_story().main()
-        m.__fields__(*self.document_fields_truncated)
-        m.text().__fields__(*self.document_text_fields_truncated)
-
-        res = self._gql_client.execute(op)
-        res = op + res
-        return res.pagination_story.list_story
-
-    def get_documents_by_limit_offset_query_extra_settings(
-        self, skip: int, take: int, query: str, extra_settings: ExtraSettings
-    ) -> Sequence[Story]:
-
-        op = Operation(Query)
-        ps: StoryPagination = op.pagination_story(
-            offset=skip,
-            limit=take,
-            extra_settings=extra_settings,
-            query=query
+            extra_settings=extra_settings if extra_settings else ExtraSettings(),
+            filter_settings=filter_settings if filter_settings else DocumentFilterSettings()
         )
         ps.list_story().list_document().__fields__(*self.document_fields_truncated)
         m = ps.list_story().main()
@@ -379,10 +440,10 @@ class TalismanAPIAdapter:
         lcf = pcf.list_concept_fact()
         lcf.__fields__(*self.concept_fact_fields)
         d = lcf.document()
-        d.__fields__(*self.concept_fact_document_fields)
+        d.__fields__(*self.document_fields_extended)
         dm = d.metadata()
-        dm.platform().__fields__(*self.concept_fact_document_platform_fields)
-        dm.account().__fields__(*self.concept_fact_document_account_fields)
+        dm.platform().__fields__(*self.document_platform_fields)
+        dm.account().__fields__(*self.document_account_fields)
 
         res = self._gql_client.execute(op)
         res = op + res
@@ -407,39 +468,64 @@ class TalismanAPIAdapter:
         return res.concept_link
 
     def get_all_concepts(
-        self, query: str = None, filter_settings: ConceptFilterSettings = None,
+        self, filter_settings: ConceptFilterSettings = None,
         direction: SortDirection = 'descending', sort_field: ConceptSorting = 'score',
         with_aliases: bool = False, with_properties: bool = False, with_links: bool = False,
-        with_link_properties: bool = False
+        with_link_properties: bool = False, with_facts: bool = False
     ) -> Iterable[Concept]:
-        concepts: Iterable = [None]
-        i: int = 0
         if not filter_settings:
             filter_settings = ConceptFilterSettings()
+        total = self.get_concept_count(filter_settings=filter_settings)
+
+        if total > self.kb_iterator_config.max_total_count:
+            had_creation_date = hasattr(filter_settings, 'creation_date')
+            old_timestamp_interval = None
+            if had_creation_date:
+                old_timestamp_interval = copy(filter_settings.creation_date)
+            start: int = getattr(old_timestamp_interval, 'start', self.kb_iterator_config.earliest_created_time)
+            end: int = getattr(old_timestamp_interval, 'end', int(time()))
+            middle: int = (end + start) // 2
+
+            for start, end in (start, middle), (middle, end):
+                filter_settings.creation_date = TimestampInterval(start=start, end=end)
+                for c in self.get_all_concepts(filter_settings=filter_settings,
+                                               direction=direction, sort_field=sort_field, with_aliases=with_aliases,
+                                               with_properties=with_properties, with_links=with_links,
+                                               with_link_properties=with_link_properties, with_facts=with_facts):
+                    yield c
+
+            if had_creation_date:
+                filter_settings.creation_date = old_timestamp_interval
+            else:
+                delattr(filter_settings, 'creation_date')
+            return
+        elif not total:
+            return
+
+        concepts: Iterable = [None]
+        i: int = 0
         while concepts:
             concepts = self.get_concepts(
-                skip=i * 20, take=20, query=query, filter_settings=filter_settings,
+                skip=i * self._limit, take=self._limit, filter_settings=filter_settings,
                 direction=direction, sort_field=sort_field, with_aliases=with_aliases,
-                with_properties=with_properties, with_links=with_links, with_link_properties=with_link_properties
+                with_properties=with_properties, with_links=with_links,
+                with_link_properties=with_link_properties, with_facts=with_facts
             )
             for c in concepts:
                 yield c
             i += 1
 
     def get_concepts(
-        self, skip: int = 0, take: int = 20, query: str = None,
-        filter_settings: ConceptFilterSettings = None,
+        self, skip: int = 0, take: Optional[int] = None, filter_settings: ConceptFilterSettings = None,
         direction: SortDirection = 'descending', sort_field: ConceptSorting = 'score',
         with_aliases: bool = False, with_properties: bool = False, with_links: bool = False,
-        with_link_properties: bool = False
+        with_link_properties: bool = False, with_facts: bool = False
     ) -> Sequence[Concept]:
+        take = self.get_take_value(take)
         pagination_concept_kwargs = dict()
         if not filter_settings:
             filter_settings = ConceptFilterSettings()
 
-        if query:
-            pagination_concept_kwargs['query'] = query
-
         op = Operation(Query)
         cp: ConceptPagination = op.pagination_concept(
             limit=take,
@@ -450,84 +536,93 @@ class TalismanAPIAdapter:
             **pagination_concept_kwargs
         )
         self._configure_output_concept_fields(
-            cp.list_concept(), with_aliases=with_aliases,
-            with_properties=with_properties, with_links=with_links, with_link_properties=with_link_properties
+            cp.list_concept(), with_aliases=with_aliases, with_properties=with_properties,
+            with_links=with_links, with_link_properties=with_link_properties, with_facts=with_facts
         )
         res = self._gql_client.execute(op)
         res = op + res
         return res.pagination_concept.list_concept
 
     def get_concepts_by_limit_offset_filter_settings(
-        self, skip: int, take: int, filter_settings: ConceptFilterSettings, with_aliases: bool = False
+        self, skip: int = 0, take: Optional[int] = None, filter_settings: Optional[ConceptFilterSettings] = None,
+        with_aliases: bool = False, with_facts: bool = False
     ) -> Sequence[Concept]:
-
+        take = self.get_take_value(take)
         op = Operation(Query)
         cp: ConceptPagination = op.pagination_concept(
-            filter_settings=filter_settings,
+            filter_settings=filter_settings if filter_settings else ConceptFilterSettings(),
             offset=skip,
             limit=take
         )
-        self._configure_output_concept_fields(cp.list_concept(), with_aliases=with_aliases)
+        self._configure_output_concept_fields(cp.list_concept(), with_aliases=with_aliases, with_facts=with_facts)
         res = self._gql_client.execute(op)
         res = op + res
         return res.pagination_concept.list_concept
 
     def get_all_concept_links(
-        self, filter_settings: ConceptLinkFilterSettings = None
+        self, filter_settings: ConceptLinkFilterSettings = None, with_link_properties: bool = False
     ) -> Iterable[ConceptLink]:
+        if not filter_settings:
+            filter_settings = ConceptLinkFilterSettings()
+
+        total = self.get_concept_link_count(filter_settings=filter_settings)
+
+        if total > self.kb_iterator_config.max_total_count:
+            had_creation_date = hasattr(filter_settings, 'creation_date')
+            old_timestamp_interval = None
+            if had_creation_date:
+                old_timestamp_interval = copy(filter_settings.creation_date)
+            start: int = getattr(old_timestamp_interval, 'start', self.kb_iterator_config.earliest_created_time)
+            end: int = getattr(old_timestamp_interval, 'end', int(time()))
+            middle: int = (end + start) // 2
+
+            for start, end in (start, middle), (middle, end):
+                filter_settings.creation_date = TimestampInterval(start=start, end=end)
+                for c in self.get_all_concept_links(filter_settings=filter_settings,
+                                                    with_link_properties=with_link_properties):
+                    yield c
+
+            if had_creation_date:
+                filter_settings.creation_date = old_timestamp_interval
+            else:
+                delattr(filter_settings, 'creation_date')
+            return
+        elif not total:
+            return
+
         links: Iterable = [None]
         i: int = 0
         while links:
             links = self.get_concept_links_by_limit_offset_filter_settings(
-                skip=i * 20, take=20,
-                filter_settings=filter_settings if filter_settings else ConceptLinkFilterSettings()
+                skip=i * self._limit, take=self._limit,
+                filter_settings=filter_settings, with_link_properties=with_link_properties
             )
             for link in links:
                 yield link
             i += 1
 
     def get_concept_links_by_limit_offset_filter_settings(
-        self, skip: int = 0, take: int = 20,
-        filter_settings: ConceptLinkFilterSettings = None
+        self, skip: int = 0, take: Optional[int] = None, filter_settings: ConceptLinkFilterSettings = None,
+        with_link_properties: bool = False
     ) -> Sequence[ConceptLink]:
-
+        take = self.get_take_value(take)
         op = Operation(Query)
         pcl: ConceptLinkPagination = op.pagination_concept_link(
             filter_settings=filter_settings if filter_settings else ConceptLinkFilterSettings(),
             offset=skip,
             limit=take
         )
-        lcl = pcl.list_concept_link()
-        lcl.__fields__(*self.concept_link_fields)
-        lcl.concept_from().__fields__(*self.concept_link_concept_from_fields)
-        lcl.concept_to().__fields__(*self.concept_link_concept_to_fields)
-        lcl.concept_link_type().__fields__(*self.concept_link_type_fields_truncated)
-
+        self._configure_output_link_fields(pcl.list_concept_link(), with_link_properties=with_link_properties)
         res = self._gql_client.execute(op)
         res = op + res
         return res.pagination_concept_link.list_concept_link
 
-    def get_concepts_by_limit_offset_query(
-        self, skip: int, take: int, query: str, with_aliases: bool = False
-    ) -> Sequence[Concept]:
-
-        op = Operation(Query)
-        cp: ConceptPagination = op.pagination_concept(
-            offset=skip,
-            limit=take,
-            query=query
-        )
-        self._configure_output_concept_fields(cp.list_concept(), with_aliases=with_aliases)
-        res = self._gql_client.execute(op)
-        res = op + res
-        return res.pagination_concept.list_concept
-
     def get_concepts_by_type_id_with_offset(
-        self, type_id: str, skip: int, take: int, direction='descending', sort_field='systemRegistrationDate',
-        with_aliases: bool = False, with_properties: bool = False, with_links: bool = False,
-        with_link_properties: bool = False
+        self, type_id: str, skip: int, take: Optional[int] = None, direction='descending',
+        sort_field='systemRegistrationDate', with_aliases: bool = False, with_properties: bool = False,
+        with_links: bool = False, with_link_properties: bool = False, with_facts: bool = False
     ) -> ConceptPagination:
-
+        take = self.get_take_value(take)
         op = Operation(Query)
         cp: ConceptPagination = op.pagination_concept(
             filter_settings=ConceptFilterSettings(
@@ -540,17 +635,19 @@ class TalismanAPIAdapter:
         )
         cp.total()
         self._configure_output_concept_fields(
-            cp.list_concept(), with_aliases, with_properties, with_links, with_link_properties
+            cp.list_concept(), with_aliases=with_aliases, with_properties=with_properties, with_links=with_links,
+            with_link_properties=with_link_properties, with_facts=with_facts
         )
         res = self._gql_client.execute(op)
         res = op + res
         return res.pagination_concept
 
     def get_concepts_by_type_id_with_offset_with_markers(
-        self, type_id: str, skip: int, take: int, markers: List[str], direction='descending',
-        sort_field='systemRegistrationDate', with_aliases: bool = False
+        self, type_id: str, skip: int = 0, take: Optional[int] = None, markers: Optional[List[str]] = None,
+        direction='descending', sort_field='systemRegistrationDate', with_aliases: bool = False,
+        with_facts: bool = False
     ) -> ConceptPagination:
-
+        take = self.get_take_value(take)
         op = Operation(Query)
         cp: ConceptPagination = op.pagination_concept(
             filter_settings=ConceptFilterSettings(
@@ -563,13 +660,13 @@ class TalismanAPIAdapter:
             sort_field=sort_field
         )
         cp.total()
-        self._configure_output_concept_fields(cp.list_concept(), with_aliases=with_aliases)
+        self._configure_output_concept_fields(cp.list_concept(), with_aliases=with_aliases, with_facts=with_facts)
         res = self._gql_client.execute(op)
         res = op + res
         return res.pagination_concept
 
     def get_concepts_by_name(
-        self, name: str, type_id: Optional[str] = None, with_aliases: bool = False
+        self, name: str, type_id: Optional[str] = None, with_aliases: bool = False, with_facts: bool = False
     ) -> Sequence[Concept]:
 
         op = Operation(Query)
@@ -585,13 +682,13 @@ class TalismanAPIAdapter:
         cp: ConceptPagination = op.pagination_concept(
             filter_settings=concept_filter_settings
         )
-        self._configure_output_concept_fields(cp.list_concept(), with_aliases=with_aliases)
+        self._configure_output_concept_fields(cp.list_concept(), with_aliases=with_aliases, with_facts=with_facts)
         res = self._gql_client.execute(op)
         res = op + res
         return res.pagination_concept.list_concept
 
     def get_concepts_by_near_name(
-        self, name: str, type_id: Optional[str] = None, with_aliases: bool = False
+        self, name: str, type_id: Optional[str] = None, with_aliases: bool = False, with_facts: bool = False
     ) -> Sequence[Concept]:
 
         op = Operation(Query)
@@ -607,13 +704,14 @@ class TalismanAPIAdapter:
         cp: ConceptPagination = op.pagination_concept(
             filter_settings=concept_filter_settings
         )
-        self._configure_output_concept_fields(cp.list_concept(), with_aliases=with_aliases)
+        self._configure_output_concept_fields(cp.list_concept(), with_aliases=with_aliases, with_facts=with_facts)
         res = self._gql_client.execute(op)
         res = op + res
         return res.pagination_concept.list_concept
 
     def get_concepts_by_property_name(
-        self, property_type_id: str, string_filter: str, property_type: str = 'concept', with_aliases: bool = False
+        self, property_type_id: str, string_filter: str, property_type: str = 'concept', with_aliases: bool = False,
+        with_facts: bool = False
     ) -> Sequence[Concept]:
 
         op = Operation(Query)
@@ -628,7 +726,7 @@ class TalismanAPIAdapter:
                 )]
             )
         )
-        self._configure_output_concept_fields(cp.list_concept(), with_aliases=with_aliases)
+        self._configure_output_concept_fields(cp.list_concept(), with_aliases=with_aliases, with_facts=with_facts)
         res = self._gql_client.execute(op)
         res = op + res
         return res.pagination_concept.list_concept
@@ -652,7 +750,7 @@ class TalismanAPIAdapter:
         res = op + res  # type: Query
         return res.concept.pagination_concept_property.list_concept_property
 
-    def get_concept_links(self, concept_id: str) -> Sequence[ConceptLink]:
+    def get_concept_links(self, concept_id: str, with_link_properties: bool = False) -> Sequence[ConceptLink]:
         op = Operation(Query)
 
         concept: Concept = op.concept(
@@ -663,17 +761,14 @@ class TalismanAPIAdapter:
             limit=10000,
             filter_settings=ConceptLinkFilterSettings()
         )
-        lcl = pcl.list_concept_link()
-        lcl.__fields__(*self.concept_link_fields)
-        lcl.concept_from().__fields__(*self.concept_link_concept_from_fields)
-        lcl.concept_to().__fields__(*self.concept_link_concept_to_fields)
-        lcl.concept_link_type().__fields__(*self.concept_link_type_fields_truncated)
-
+        self._configure_output_link_fields(pcl.list_concept_link(), with_link_properties=with_link_properties)
         res = self._gql_client.execute(op)
         res = op + res  # type: Query
         return res.concept.pagination_concept_link.list_concept_link
 
-    def get_concept_links_concept(self, concept_id: str, link_type_id: str) -> Sequence[ConceptLink]:
+    def get_concept_links_concept(
+            self, concept_id: str, link_type_id: str, with_link_properties: bool = False
+    ) -> Sequence[ConceptLink]:
         op = Operation(Query)
 
         concept = op.concept(
@@ -683,15 +778,10 @@ class TalismanAPIAdapter:
             offset=0,
             limit=10000,
             filter_settings=ConceptLinkFilterSettings(
-                concept_link_type=link_type_id
+                concept_link_type=[link_type_id]
             )
         )
-        lcl = pcl.list_concept_link()
-        lcl.__fields__(*self.concept_link_fields)
-        lcl.concept_from().__fields__(*self.concept_link_concept_from_fields)
-        lcl.concept_to().__fields__(*self.concept_link_concept_to_fields)
-        lcl.concept_link_type().__fields__(*self.concept_link_type_fields_truncated)
-
+        self._configure_output_link_fields(pcl.list_concept_link(), with_link_properties=with_link_properties)
         res = self._gql_client.execute(op)
         res = op + res  # type: Query
         return res.concept.pagination_concept_link.list_concept_link
@@ -1155,6 +1245,18 @@ class TalismanAPIAdapter:
 
         return relation
 
+    def delete_concept(self, concept_id: str) -> bool:
+
+        op = Operation(Mutation)
+        dc: Concept = op.delete_concept(
+            id=concept_id
+        )
+        dc.__fields__('is_success')
+        res = self._gql_client.execute(op)
+        res = op + res
+
+        return res.delete_concept.is_success
+
     def delete_concept_link(self, link_id: str) -> bool:
 
         op = Operation(Mutation)
@@ -1231,7 +1333,8 @@ class TalismanAPIAdapter:
 
     # region Crawlers methods
 
-    def get_crawler_start_urls(self, take: int) -> Sequence[Crawler]:
+    def get_crawler_start_urls(self, take: Optional[int] = None) -> Sequence[Crawler]:
+        take = self.get_take_value(take)
         op = Operation(CrQuery)
         pc: CrawlerPagination = op.pagination_crawler(
             limit=take
